#ifndef RE__H
#define RE__H
#include <stdio.h>
#include "ieee-nan.h"

class RE
{
  double v;
public:
  // Construction and conversion operations
  RE () { v = NaN; }
  RE (double a) : v(a) { }
  RE (const RE& a) { v = a.v; }
  ~RE () { }
  void operator =( const RE& a ) { this->v = a.v; }
  // Fundamental comparison operators
  friend bool operator ==( const RE& a, const RE& b );
  friend bool operator <=( const RE& a, const RE& b );
  friend long IsZero( const RE& a);
  friend long IsOne ( const RE& a);
  // Fundamental arithmetic operators
  friend RE operator -( const RE& a );
  friend RE inv  ( const RE&  a);
  friend RE sqrt ( const RE&  a);
  RE& operator +=( const RE& a );
  RE& operator *=( const RE& a );
  RE& operator -=( const RE& a ) { *this += -a; return *this; }
  RE& operator /=( const RE& a ) { *this *= inv(a); return *this; }
  long log2() const; // returns the precision of a ``0''
  // Fundamental output functions
  friend void operator <<( double& z, const RE& a );
  friend void operator <<( RE& z, const char* a );
  friend void operator <<( char*& s, const RE& r );
  
  // Definitions of fast additional methods
  friend RE operator +( const RE& a, const RE& b );
  friend RE operator -( const RE& a, const RE& b );
  friend RE operator *( const RE& a, const RE& b );
  friend RE operator /( const RE& a, const RE& b );
};

// Fundamental comparison operators
inline bool operator ==( const RE& a, const RE& b ) { return a.v == b.v; }
inline bool operator <=( const RE& a, const RE& b ) { return a.v <= b.v; }
inline long IsZero ( const RE& a ) { return a.v == 0.; }
inline long IsOne  ( const RE& a ) { return a.v == 1.; }
// Fundamental arithmetic operators
inline RE operator - ( const RE& a ) { RE r; r.v = - a.v; return r; }
inline RE inv  ( const RE& a ) { RE r; r.v = 1./a.v; return r; }
inline RE sqrt ( const RE& a ) { RE r; r.v = sqrt(a.v); return r; }
inline RE& RE::operator += ( const RE& a ) { v += a.v; return *this; }
inline RE& RE::operator *= ( const RE& a ) { v *= a.v; return *this; }
inline long RE::log2() const { return static_cast<long>(-log(v)/log(2.)); }
// Fundamental output functions
inline void operator <<( double& z, const RE& a ) { z = a.v; }
inline void operator <<( RE& z, const char* a ) { sscanf(a, "%lf", &z.v); }
// [re-io.c] void operator <<( char*& s, const RE& r );

// -------- Redefinitions of additional methods --------
inline RE operator + (const RE& a, const RE& b) {RE r;r.v=a.v+b.v;return r;}
inline RE operator - (const RE& a, const RE& b) {RE r;r.v=a.v-b.v;return r;}
inline RE operator * (const RE& a, const RE& b) {RE r;r.v=a.v*b.v;return r;}
inline RE operator / (const RE& a, const RE& b) {RE r;r.v=a.v/b.v;return r;}

// -------- Definitions from fundamental methods --------
// Standard arithmetic operators (as functions)
inline void add(RE& z, const RE& a, const RE& b) { z = a + b; }
inline void sub(RE& z, const RE& a, const RE& b) { z = a - b; }
inline void mul(RE& z, const RE& a, const RE& b) { z = a * b; }
// Other operators
inline RE sqr ( const RE& a ) { return a * a; }
inline bool operator !=( const RE& a, const RE& b ) { return !(a == b); }

#endif
