#ifndef H_REAL
#define H_REAL

#include "useproto.h"
#include <math.h>
#include <iostream.h>
#include <malloc.h>
#include "ieee-nan.h"

#if !defined(bool) && defined(__SUNPRO_CC)
#define bool bool
typedef int bool;
const bool false = 0;
const bool true  = 1;
#endif

enum FMT { F_short, F_medium, F_long, F_pari };

class RE_IO
{
  static FMT fmt;
  static void output(FMT f) { fmt = f; }
  friend ostream& IOpari  (ostream&);
  friend ostream& IOlong  (ostream&);
  friend ostream& IOmedium(ostream&);
  friend ostream& IOshort (ostream&);
public:
  static FMT  output(void)  { return fmt; }
};

inline ostream& IOpari  (ostream& o) { RE_IO::fmt = F_pari;   return o; }
inline ostream& IOlong  (ostream& o) { RE_IO::fmt = F_long;   return o; }
inline ostream& IOmedium(ostream& o) { RE_IO::fmt = F_medium; return o; }
inline ostream& IOshort (ostream& o) { RE_IO::fmt = F_short;  return o; }

/**** Pari/GP ****/
#ifdef mplib_pari
#ifdef mpopt_fast
#include <re-pari-fast.h>
#endif
#include <re-pari.h>
#define __RE_GARBAGE_BEGIN  long avma_garbage = avma;
#define __RE_GARBAGE_END    avma = avma_garbage;
#endif

/**** GNU's MP ****/
#ifdef mplib_gmp
#error pas implante
#endif

/**** Shoup's NTL ****/
#ifdef mplib_ntl
#include <re-ntl.h>
#define __RE_GARBAGE_BEGIN
#define __RE_GARBAGE_END
#endif

/**** Double ****/
#ifdef mplib_double
#include <re-double.h>
#define __RE_GARBAGE_BEGIN
#define __RE_GARBAGE_END
#endif

inline bool operator >=( const RE& a, const RE& b ) { return  (b <= a); }
inline bool operator < ( const RE& a, const RE& b ) { return !(b <= a); }
inline bool operator > ( const RE& a, const RE& b ) { return !(a <= b); }

inline bool NearZero(const RE& a) { return 2 * log2(a) < -precision(a); }

inline ostream& operator<<(ostream& s, const RE& a)
  { char* str = a; s << str; free(str); return s; }

#endif
