#include "algdep.h"

#if 0 && (defined(mplib_pari) || defined(mplib_ntl))
#  include <unistd.h>
#  include <sys/wait.h>
#  define __INIT_FORK if (fork()) return;
#  define __EXIT_FORK _exit(0);
#  define __WAIT_FORK wait(0);
#  include "ad-signame.h" /* Re-defines __WAIT_FORK and maybe __INIT_FORK */
#else
#  define __INIT_FORK
#  define __EXIT_FORK
#  define __WAIT_FORK
#endif

#if   defined(mplib_ntl)
#include <LLL.h>
#include <ZZX.h>
typedef void (ALGalgo)(matrix(ZZ)& B);
static ALGalgo ALG_LLL;
static ALGalgo ALG_LLL_FP;
static ALGalgo ALG_BKZxFP;
static ALGalgo ALG_BKZ_FP;
static ALGalgo ALG_LLL_XD;
static ALGalgo ALG_BKZxXD;
static ALGalgo ALG_BKZ_XD;
typedef void (LLLf)(matrix(ZZ)& B, matrix(ZZ)& U);
static LLLf LLLFP;
static LLLf LLLXD;
static void minpol_algdep(LLLf fun, const CX& c,
    const ulong deg, const int ziring, const ulong digitprec);
#elif defined(mplib_pari)
typedef void (ALGalgo)(GEN& x, ulong prec);
static ALGalgo ALG_LLL;
static ALGalgo ALG_Hastad;
#else
typedef void (ALGalgo)(void);
static void ALG_LLL(void) { }
#endif
static void minpol_algo(ALGalgo ALGproc, const CX& c,
    const ulong deg, const int ziring, const ulong digitprec);


void
polynome_minimal (const CX& c,  const ulong deg, const bool ziring, ulong algo)
{
  ulong cdg = cdigitprec();
  ulong lll = cdg - 1 - (long)(2*log(deg+0.)/log(10.));
#if   defined(mplib_ntl)
  if (!algo) algo = 0x0002;
  if (algo&0x0002) minpol_algdep(   LLLFP, c, deg, ziring, lll);    __WAIT_FORK
  if (algo&0x0010) minpol_algo(ALG_LLL_FP, c, deg, ziring, lll);    __WAIT_FORK
  if (algo&0x0020) minpol_algo(ALG_LLL_FP, c, deg, ziring, lll-1);  __WAIT_FORK
  if (algo&0x0040) minpol_algo(ALG_BKZxFP, c, deg, ziring, lll);    __WAIT_FORK
  if (algo&0x0080) minpol_algo(ALG_BKZ_FP, c, deg, ziring, lll);    __WAIT_FORK
  if (algo&0x0004) minpol_algdep(   LLLXD, c, deg, ziring, lll);    __WAIT_FORK
  if (algo&0x0100) minpol_algo(ALG_LLL_XD, c, deg, ziring, lll);    __WAIT_FORK
  if (algo&0x0200) minpol_algo(ALG_LLL_XD, c, deg, ziring, lll-1);  __WAIT_FORK
  if (algo&0x0400) minpol_algo(ALG_BKZxXD, c, deg, ziring, lll);    __WAIT_FORK
  if (algo&0x0800) minpol_algo(ALG_BKZ_XD, c, deg, ziring, lll);    __WAIT_FORK
  if (algo&0x0001) minpol_algo(ALG_LLL,    c, deg, ziring, lll);    __WAIT_FORK
#elif defined(mplib_pari)
  if (!algo) algo = 0x0003;
  if (algo&0x0001) minpol_algo(ALG_Hastad, c, deg, ziring, cdg);    __WAIT_FORK
  if (algo&0x0002) minpol_algo(ALG_LLL,    c, deg, ziring, lll);    __WAIT_FORK
#else
  minpol_algo(ALG_LLL,    c, deg, ziring, lll);
  cerr << "TODO: valeur algbrique\n";
#endif
}

/**************************************************************************/

#if   defined(mplib_ntl)
typedef struct { ZZX x, y; ZZ norm; ulong bad; } ZIX;

static
ostream& ZIXoutPari(ostream& s, const ZIX& a)
{
  ulong n, i;
  ZZ   zx, zy;
  n = a.x.rep.length(); i = a.y.rep.length(); if (n<i) n = i;
  for (i = n-1; ; i--) {
    long star = 1;
    zx = coeff(a.x, i);
    zy = coeff(a.y, i);
    if (IsZero(zx)&& IsZero(zy)) continue;
    if (IsZero(zy)) {
      if (i < n-1) {
        if (zx < 0) s << " - ";
        else        s << " + ";
        if (zx < 0) negate(zx, zx);
      }
      if (IsOne(zx) && (i>0) ) star = 0;
      else                     s << zx;
      goto monome;
    }
    if (IsZero(zx)) {
      if (i < n-1) {
        if (zy < 0) s << " - ";
        else        s << " + ";
        if (zy < 0) negate(zy, zy);
      }
      if (IsOne(zy)) s << "I";
      else           s << zy << "*I";
      goto monome;
    }
    if (i < n-1) s << " + ";
    s << '(' << zx;
    if (zy < 0) s << " - ";
    else        s << " + ";
    if (zy < 0) negate(zy, zy);
    if (IsOne(zy)) s << "I)";
    else           s << zy << "*I)";
  monome:
    if (i == 0) break;
    if (star) s << '*';
    s << 'x';
    if (i == 1) continue;
    s << '^' << i;
  }
  if (a.bad >>1) s << " [" << (a.bad >> 1) << ']';
  if (a.bad & 1) s << " *** " << a.norm;
  return s;
}

static int ZIXoutFlag = 0;
static ostream& operator<<(ostream& s, const ZIX& a) //
{
  if (ZIXoutFlag == 0) return ZIXoutPari(s, a);
  else return s << "Real " << a.x << "Imag " << a.y;
}

static long operator==(const ZIX& a, const ZIX& b) //
{
  return (a.x == b.x) && (a.y == b.y);
}

static matrix(ZZ) //
algdep_lllmat(const CX *x, const ulong deg,
              const int ziring, const ulong digitprec)
{
  ulong dim = deg;
  dim++; if (ziring) dim *= 2;
  matrix(ZZ) B;
  B.SetDims(dim, dim+2);
  RE X, T;
  power(T, 10, digitprec);
  ulong i, j;
  for (i = 0; i < dim; i++) {
    for (j = 0; j < dim; j++) {
      if (i==j) set(B[i][j]);
      else    clear(B[i][j]);
    }
    mul(X, x[i].Re(), T); B[i][dim+0] << X;
    mul(X, x[i].Im(), T); B[i][dim+1] << X;
  }
  return B;
}
#endif

/**************************************************************************/

#if   defined(mplib_pari)
static GEN
algdep_vector(const CX& c, const ulong deg, const int ziring)
{
  /* On s'inspire de pari:bibli1.c (algdep0) */
  /* Mais on rajoute I fois le vecteur si ziring == 1 */
  GEN x, y;
  ulong i;
  x = c.genvalue ();
  y = cgetg(deg+2,t_COL);
  y[1] = un;
  for (i = 0; i < deg; i++) y[i+2] = lmul((GEN)y[i+1], x);
  if (ziring) y = concat(y, gmul(y, gi));
  return y;
}

#elif defined(mplib_ntl)
static CX*
algdep_vector(const CX& c, const ulong deg, const int ziring)
{
  CX *y;
  ulong i;
  if (ziring) {
    y = (CX*) malloc(2*(deg+1)*sizeof(CX));
    new (& y[0]) CX(1,0);
    new (& y[1]) CX(0,1);
    for (i = 0; i < deg; i++) {
      new (& y[2*i+2]) CX(y[2*i+0]); y[2*i+2] *= c;
      new (& y[2*i+3]) CX(y[2*i+1]); y[2*i+3] *= c;
    }
  }
  else {
    y = (CX*) malloc((deg+1)*sizeof(CX));
    new (& y[0]) CX(1);
    for (i = 0; i < deg; i++) {
      new (& y[i+1]) CX(y[i]); y[i+1] *= c;
    }
  }
  return y;
}
#endif

#if   defined(mplib_pari)
static GEN
algdep_poly(GEN x, int ziring)
{
  /* On s'inspire de pari:bibli1.c (algdep0) */
  ulong deg = lg(x) - 1;
  if (ziring) deg /= 2;
  deg--;
  GEN y = cgetg(deg+3,t_POL); 
  y[1] = evalsigne(1) | evalvarn(0);
  ulong i, j;
  for (i=j=0; i<=deg; i++)
  {
    y[i+2] = lcopy((GEN)x[i+1]);
    if (!gcmp0((GEN)x[i+1])) j=i;
  }
  setlgef(y,j+3);
  GEN z;
  if (!ziring) goto exit;
  z = y; x += deg + 1;
  y = cgetg(deg+3,t_POL); 
  y[1] = evalsigne(1) | evalvarn(0);
  for (i=j=0; i<=deg; i++)
  {
    y[i+2] = lcopy((GEN)x[i+1]);
    if (!gcmp0((GEN)x[i+1])) j=i;
  }
  setlgef(y,j+3);
  y = gmul(y,gi);
  y = gadd(z,y);
 exit:
  /* Division par x^d ventuellement */
  z=cgetg(4,t_POL);z[1]=evalsigne(1)|evallgef(4)|evalvarn(0);z[2]=zero;z[3]=un;
  while (gcmp0((GEN)y[2])) y = gdiv(y, z);
  /* Division par le contenu */
  y = gdiv(y,content(y));
  if (t_POL != typ(y)) return gzero;
  /* Canonicisation du coeff dominant */
  if (gcmp( gabs(greal((GEN)y[lgef(y)-1]),0)
          , gabs(gimag((GEN)y[lgef(y)-1]),0)
          ) < 0 ) y = gmul(y,gi);
  if (signe(greal((GEN)y[lgef(y)-1]))<0) y = gneg(y);
  return y;
}

#elif defined(mplib_ntl)
static ZIX
algdep_poly(const matrix(ZZ)& B, const int ziring, const long row)
{
  ZIX p;
  long i;
  if (ziring) {
    for (i = 0; i < B.NumCols()-2; i++) {
      SetCoeff(p.x, long(i/2), B[row][i]); i++;
      SetCoeff(p.y, long(i/2), B[row][i]);
    }
  }
  else {
    for (i = 0; i < B.NumCols()-2; i++)
      SetCoeff(p.x, i, B[row][i]);
  }
  InnerProduct(p.norm, B[row], B[row]);
  if (row < B.NumRows()-ziring-1) {
    ZZ n;
    InnerProduct(n, B[row+1+ziring], B[row+1+ziring]);
    div(n, n, p.norm);
    p.bad = ( n < 100 ) + row >> 1;
  }
  else p.bad = 1 + row >> 1;
  if (deg(p.x)<0 && deg(p.y)<0) return p;
  /* Division par x^d ventuellement */
  i = 0; while (IsZero(coeff(p.x,i))&&IsZero(coeff(p.y,i))) i++;
  RightShift(p.x, p.x, i);
  RightShift(p.y, p.y, i);
  /* Division par le contenu */
  ZZ cx, cy;
  content(cx, p.x);
  content(cy, p.y);
  ZZ c;
  GCD(c, cx, cy);
  divide(p.x, p.x, c);
  divide(p.y, p.y, c);
  /* Canonicisation du coeff dominant */
  i = deg(p.x); if (i<deg(p.y)) i = deg(p.y);
  abs(cx, coeff(p.x,i));
  abs(cy, coeff(p.y,i));
  if (cx < cy) {
    ZZX t = p.y;
    p.y = p.x;
    negate(p.x, t);
  }
  if (coeff(p.x,i) < 0) {
    negate(p.x, p.x);
    negate(p.y, p.y);
  }
  /* Racines triviales => a virer ? TODO ? */
  return p;
}

static ZIX
algdep_poly(const matrix(ZZ)& B, const int ziring)
{
  ZIX p;
  long row = -1;
  while (++row < B.NumRows()) {
    p = algdep_poly(B, ziring, row);
    if (!(p.bad&1)) break;
  }
  if (p.bad&1)
    p = algdep_poly(B, ziring, 0);
  return p;
}
#endif

#define TIMER_RESULT(t) \
  printf("en %4ld.%03ld s.  ", t/1000, t%1000); \
  if      (t > 2*3600*1000) cout << t/(3600*1000) << " heures\n"; \
  else if (t > 2*60*1000)   cout <<   t/(60*1000) << " minutes\n"; \
  cout << flush

#if   defined(mplib_ntl)
static void
ALG_LLL(matrix(ZZ)& B)
{
  ZZ n = B[0][B.NumCols()-2];
  ulong digitprec = 0;
  while (n > 1) { digitprec++; div(n, n, 10); }
  cout << "LLL    [" << digitprec << "]..." << flush;
  timer2();
  ZZ det;
  LLL(det, B);
  long t = timer2();
  TIMER_RESULT(t);
}

static void
ALG_LLL_FP(matrix(ZZ)& B)
{
  ZZ n = B[0][B.NumCols()-2];
  ulong digitprec = 0;
  while (n > 1) { digitprec++; div(n, n, 10); }
  cout << "LLL FP [" << digitprec << "]..." << flush;
  timer2();
  LLL_FP(B);
  long t = timer2();
  TIMER_RESULT(t);
}

static void
ALG_LLL_XD(matrix(ZZ)& B)
{
  ZZ n = B[0][B.NumCols()-2];
  ulong digitprec = 0;
  while (n > 1) { digitprec++; div(n, n, 10); }
  cout << "LLL XD [" << digitprec << "]..." << flush;
  timer2();
  LLL_XD(B);
  long t = timer2();
  TIMER_RESULT(t);
}

static void
ALG_BKZ_FP(matrix(ZZ)& B)
{
  ZZ n = B[0][B.NumCols()-2];
  ulong digitprec = 0;
  while (n > 1) { digitprec++; div(n, n, 10); }
  cout << "BKZ FP [" << digitprec << "]..." << flush;
  timer2();
  BKZ_FP(B);
  long t = timer2();
  TIMER_RESULT(t);
}

static void
ALG_BKZ_XD(matrix(ZZ)& B)
{
  ZZ n = B[0][B.NumCols()-2];
  ulong digitprec = 0;
  while (n > 1) { digitprec++; div(n, n, 10); }
  cout << "BKZ XD [" << digitprec << "]..." << flush;
  timer2();
  BKZ_XD(B);
  long t = timer2();
  TIMER_RESULT(t);
}

static void
ALG_BKZxFP(matrix(ZZ)& B)
{
  ZZ n = B[0][B.NumCols()-2];
  ulong digitprec = 0;
  while (n > 1) { digitprec++; div(n, n, 10); }
  cout << "BKZ*FP [" << digitprec << "]..." << flush;
  timer2();
  BKZ_FP(B,0.99,B.NumRows(),1);
  long t = timer2();
  TIMER_RESULT(t);
}

static void
ALG_BKZxXD(matrix(ZZ)& B)
{
  ZZ n = B[0][B.NumCols()-2];
  ulong digitprec = 0;
  while (n > 1) { digitprec++; div(n, n, 10); }
  cout << "BKZ*XD [" << digitprec << "]..." << flush;
  timer2();
  BKZ_XD(B,0.99,B.NumRows(),1);
  long t = timer2();
  TIMER_RESULT(t);
}
#elif defined(mplib_pari)
static void
ALG_Hastad(GEN& x, ulong digitprec)
{
  /* L'appel  algdep peut planter,            */
  /* avec un "precision loss in truncation"    */
  /* Prcision des rels Pari */
  ulong wordprec = 3 + (ulong) (digitprec/L2SL10/8/sizeof(long));
  if ( (int)(log(wordprec+0.)/log(10.)) < (int)(log(digitprec+0.)/log(10.)) )
    cout << "Hastad  <" << wordprec << ">..." << flush;
  else
    cout << "Hastad <"  << wordprec << ">..." << flush;
  timer2();
  x = lindep(x, wordprec);
  long t = timer2();
  TIMER_RESULT(t);
}

static void
ALG_LLL(GEN& x, ulong digitprec)
{
  /* L'appel  algdep2 rame...                 */
  cout << "LLL    [" << digitprec << "]..." << flush;
  timer2();
  /* Prcision en digits dcimaux              */
  x = lindep2(x, digitprec);
  long t = timer2();
  TIMER_RESULT(t);
}

#endif

static void
minpol_algo(ALGalgo ALGproc, const CX& c,
  const ulong deg, const int ziring, const ulong digitprec)
{
#if   defined(mplib_ntl)
  __INIT_FORK
  CX* x = algdep_vector(c, deg, ziring);
  matrix(ZZ) B = algdep_lllmat(x, deg, ziring, digitprec);
  ALGproc(B);
  cout << algdep_poly(B, ziring) << "\n";
  fflush(stdout);
  __EXIT_FORK
#elif defined(mplib_pari)
  __INIT_FORK
  GEN x = algdep_vector(c, deg, ziring);
  ALGproc(x, digitprec);
  output(algdep_poly(x, ziring));
  fflush(stdout);
  __EXIT_FORK
#endif
}

#if   defined(mplib_ntl)
static ZZ
getsize(matrix(ZZ) B, ulong deg, int ziring)
{
  ZZ x, y;
  sqr(x, B[0][(ziring+1)*(deg+1)]);
  sqr(y, B[0][(ziring+1)*(deg+1)+1]);
  add(x, x, y);
  return x;
}

static void LLLFP(matrix(ZZ)& B, matrix(ZZ)& U) { LLL_FP(B,U); }
static void LLLXD(matrix(ZZ)& B, matrix(ZZ)& U) { LLL_XD(B,U); }
static void
minpol_algdep(LLLf fun, const CX& c,
  const ulong deg, const int ziring, const ulong digitprec)
{
  __INIT_FORK
  CX*  x = algdep_vector(c, deg, ziring);
  ulong dprec = 10+deg; if (digitprec < dprec) dprec = digitprec;
  matrix(ZZ) V; ident(V, (ziring+1)*(deg+1));
  if (fun == LLLFP) cout << "ALGFP " << flush; else
  if (fun == LLLXD) cout << "ALGXD " << flush; else
                    cout << "ALGDEP" << flush;
  timer2();
  matrix(ZZ) B = algdep_lllmat(x, deg, ziring, dprec);
  matrix(ZZ) U;
  ZIX bres;
  for (;;) {
    fun(B,U);
    mul(V, U,V);
    bres = algdep_poly(B, ziring);
    cout << "." << flush;
    if (dprec == digitprec) break;
    dprec += 10+deg; if (digitprec < dprec) dprec = digitprec;
    ZZ bsize = getsize(B, deg, ziring);
    B = algdep_lllmat(x, deg, ziring, dprec);
    mul(B, V,B);
    mul(bsize, bsize, 10);
    if ( bsize > getsize(B, deg, ziring) ) break;
  }
  long t = timer2();
  cout << '[' << dprec << "] ";
  TIMER_RESULT(t);
  cout << bres << "\n" << flush;
  __EXIT_FORK
}
#endif
