#include "fenetre.h"
#include <strings.h>

const char*
XEventType (XEvent ev)
{
  switch (ev.type) {
    case KeyPress:         return "KeyPress"; break;
    case KeyRelease:       return "KeyRelease"; break;
    case ButtonPress:      return "ButtonPress"; break;
    case ButtonRelease:    return "ButtonRelease"; break;
    case MotionNotify:     return "MotionNotify"; break;
    case EnterNotify:      return "EnterNotify"; break;
    case LeaveNotify:      return "LeaveNotify"; break;
    case FocusIn:          return "FocusIn"; break;
    case FocusOut:         return "FocusOut"; break;
    case KeymapNotify:     return "KeymapNotify"; break;
    case Expose:           return "Expose"; break;
    case GraphicsExpose:   return "GraphicsExpose"; break;
    case NoExpose:         return "NoExpose"; break;
    case VisibilityNotify: return "VisibilityNotify"; break;
    case CreateNotify:     return "CreateNotify"; break;
    case DestroyNotify:    return "DestroyNotify"; break;
    case UnmapNotify:      return "UnmapNotify"; break;
    case MapNotify:        return "MapNotify"; break;
    case MapRequest:       return "MapRequest"; break;
    case ReparentNotify:   return "ReparentNotify"; break;
    case ConfigureNotify:  return "ConfigureNotify"; break;
    case ConfigureRequest: return "ConfigureRequest"; break;
    case GravityNotify:    return "GravityNotify"; break;
    case ResizeRequest:    return "ResizeRequest"; break;
    case CirculateNotify:  return "CirculateNotify"; break;
    case CirculateRequest: return "CirculateRequest"; break;
    case PropertyNotify:   return "PropertyNotify"; break;
    case SelectionClear:   return "SelectionClear"; break;
    case SelectionRequest: return "SelectionRequest"; break;
    case SelectionNotify:  return "SelectionNotify"; break;
    case ColormapNotify:   return "ColormapNotify"; break;
    case ClientMessage:    return "ClientMessage"; break;
    case MappingNotify:    return "MappingNotify"; break;
  }
  return "UNDEFINED";
}

FENETRE::~FENETRE ()
{
  XFreePixmap    (dpy,pix);
  XFreeGC        (dpy,gc);
  XCloseDisplay  (dpy);
}

FENETRE::FENETRE (const char* winname)
{
  dpy                        = XOpenDisplay (NULL);
  if (!dpy)                  { perror ("Cannot access X server"); return; }
  scr                        = DefaultScreen (dpy);
  int xpos                   = 30;
  int ypos                   = 40;
  width                      = DisplayWidth  (dpy,scr) > 700 ? 600 : 300;
  height                     = DisplayHeight (dpy,scr) > 700 ? 600 : 300;
  unsigned int border_width  = 6;
  unsigned long border       = BlackPixel (dpy,scr);
  unsigned long background   = WhitePixel (dpy,scr);
  win                        = XCreateSimpleWindow (dpy,DefaultRootWindow (dpy),
                               xpos,ypos,width,height,
                               border_width,border,background);
  unsigned int depth         = DefaultDepth (dpy,scr);
  pix                        = XCreatePixmap (dpy,win, width,height,depth);
  gc                         = XCreateGC (dpy,pix, 0,NULL);
  XStoreName                  (dpy,win, winname);
  XSelectInput                (dpy,win,
       ExposureMask|KeyPressMask|ButtonPressMask|StructureNotifyMask);
  // affichage
  XMapWindow                  (dpy,win);
  effacefenetre               ();
  XEvent ev;
  do { XNextEvent(dpy,&ev);
  } while (ev.type != Expose);
  // Paramtres d'chelle, par dfaut
  translatex = translatey = 0.;
  scalex = scaley = 100.;
}

void
FENETRE::effacefenetre () const
{
  XSetForeground(dpy,gc,couleur(couleur_fond));
  XFillRectangle (dpy, pix, gc, 0,0, width,height);
}

void
FENETRE::dessinefenetre () const
{
  XCopyArea (dpy, pix, win, gc, 0,0, width,height, 0,0);
  XFlush (dpy);
}

void
FENETRE::winscale (const double factor)
{
  scalex *= factor;
  scaley *= factor;
}

void
FENETRE::winmove (const KeySym arrow)
{
  switch (arrow) {
    case XK_Up:    translatey += height/scaley/4; break;
    case XK_Down:  translatey -= height/scaley/4; break;
    case XK_Right: translatex -= width/scalex/4; break;
    case XK_Left:  translatex += width/scalex/4; break;
  }
}


static unsigned short
color_convert (double x)
{
  if (x >= 1.) x = 65535./65536;
  return static_cast<unsigned short> (65536 * x);
}

static XColor
XAllocNearestColor (Display* dpy, Colormap c, XColor in, short distance)
{
  static short bloc = 1024;
  // Dbut du parcours
  if (!distance && XAllocColor (dpy, c, &in)) return in;
  // Voisins  distance fixe (par blocs)...
  short limite = short (sqrt (distance));
  for (short x = -limite; x <= limite; x++)
    for (short y = -limite; y <= limite; y++)
      for (short z = -limite; z <= limite; z++)
        if (x*x + y*y + z*z == distance) {
    XColor col;
    int r = in.red;   r+=bloc*x; col.red   = (r>65535) ? 65535 : (r<0) ? 0 : r;
    int g = in.green; g+=bloc*y; col.green = (g>65535) ? 65535 : (g<0) ? 0 : g;
    int b = in.blue;  b+=bloc*z; col.blue  = (b>65535) ? 65535 : (b<0) ? 0 : b;
    if (XAllocColor (dpy, c, &col)) return col;
  }
  if (distance >= 16384) return in;
  return XAllocNearestColor (dpy, c, in, distance + 1);
}

unsigned long
FENETRE::RGBcolor (double r, double g, double b) const
{
  XColor couleur;
  couleur.red   = color_convert (r);
  couleur.green = color_convert (g);
  couleur.blue  = color_convert (b);
  //cerr<<"Wanted : "<<couleur.red<<","<<couleur.green<<","<<couleur.blue<<"\n";
  Colormap c = DefaultColormap (dpy, scr);
  couleur = XAllocNearestColor (dpy, c, couleur, 0);
  //cerr<<"Result : "<<couleur.red<<","<<couleur.green<<","<<couleur.blue<<"\n";
  return couleur.pixel;
}

void
FENETRE::coordonnees_en_pixels
  (unsigned int& x, unsigned int& y, const CX& c) const
{
  double cx, cy;
  ctodbl (cx, cy, c);
  cx -= translatex; cx *= scalex; cx = width/2 + cx;
  cy -= translatey; cy *= scaley; cy = height/2 - cy;
  if (cx < 0) x = 0; else if (cx > width)  x = width;
  else x = static_cast<unsigned int> (floor(cx));
  if (cy < 0) y = 0; else if (cy > height) y = height;
  else y = static_cast<unsigned int> (floor(cy));
}

CX
FENETRE::coordonnees_complexes
  (const unsigned int x, const unsigned int y) const
{
  double cx, cy;
  cx = x;  cx -= width/2; cx /= scalex; cx += translatex;
  cy = height/2; cy -= y; cy /= scaley; cy += translatey;
  return CX(cx, cy);
}

void
FENETRE::axes (const COULEUR c) const
{
  __RE_GARBAGE_BEGIN
  static const int marge = 10;
  XSetForeground(dpy,gc,couleur(c));
  CX hautgauche = coordonnees_complexes(marge, marge);
  CX basdroite  = coordonnees_complexes(width-marge, height-marge);
  double haut; double gauche; ctodbl (gauche, haut, hautgauche);
  double bas;  double droite; ctodbl (droite, bas , basdroite);
  // Axe vertical
  XDrawLine(dpy,pix,gc, marge/2,marge,marge/2,height-marge);
  XDrawLine(dpy,pix,gc, marge/2,marge,marge/2-3,marge+3);
  XDrawLine(dpy,pix,gc, marge/2,marge,marge/2+3,marge+3);
  double echelley = ( height - 2 * marge) / scaley;
  echelley = exp(ceil(log(echelley)/log(10.)-1)*log(10));
  if      ((haut-bas)/echelley < 2) echelley /= 5;
  else if ((haut-bas)/echelley < 4) echelley /= 2;
  {for (int i= static_cast<int> (ceil(bas/echelley)); echelley * i < haut; i++)
  {
    double coord = echelley * i;
    unsigned int x, y;
    coordonnees_en_pixels (x, y, CX(0.,coord));
    XTextItem textitem;
    textitem.delta = 0;
    textitem.font = None;
    textitem.chars = new char[100];
    snprintf(textitem.chars, 100, "%g", coord);
    textitem.nchars = strlen(textitem.chars);
    XDrawText(dpy,pix,gc,marge,y+3, &textitem, 1);
    delete[] textitem.chars;
    XDrawLine(dpy,pix,gc, 3,y,marge-3,y);
  }}
  // Axe horizontal
  XDrawLine(dpy,pix,gc, marge,marge/2,width-marge,marge/2);
  XDrawLine(dpy,pix,gc, width-marge,marge/2,width-marge-3,marge/2-3);
  XDrawLine(dpy,pix,gc, width-marge,marge/2,width-marge-3,marge/2+3);
  double echellex = ( width - 2 * marge) / scalex;
  echellex = exp(ceil(log(echellex)/log(10.)-1)*log(10));
  if      ((droite-gauche)/echellex < 2) echellex /= 5;
  else if ((droite-gauche)/echellex < 4) echellex /= 2;
  {for (int i= static_cast<int> (ceil(gauche/echellex)); echellex*i<droite; i++)
  {
    double coord = echellex * i;
    unsigned int x, y;
    coordonnees_en_pixels (x, y, CX(coord,0.));
    XTextItem textitem;
    textitem.delta = 0;
    textitem.font = None;
    textitem.chars = new char[100];
    snprintf(textitem.chars, 100, "%g", coord);
    textitem.nchars = strlen(textitem.chars);
    XDrawText(dpy,pix,gc,x-3,marge+8, &textitem, 1);
    delete[] textitem.chars;
    XDrawLine(dpy,pix,gc, x,3,x,marge-3);
  }}
  __RE_GARBAGE_END
}

void
FENETRE::ligne (const CX& a, const CX& b,
  const COULEUR c, const int largeur) const
{
  unsigned int xa, ya, xb, yb;
  coordonnees_en_pixels ( xa,ya, a );
  coordonnees_en_pixels ( xb,yb, b );
  XSetForeground(dpy,gc,couleur(c));
  if (largeur != 1) {
    XGCValues gc_values;
    gc_values.line_width = largeur;
    XChangeGC (dpy,gc, GCLineWidth, &gc_values);
  }
  XDrawLine (dpy,pix,gc, xa,ya, xb,yb);
  if (largeur != 1) {
    XGCValues gc_values;
    gc_values.line_width = 1;
    XChangeGC (dpy,gc, GCLineWidth, &gc_values);
  }
}

void
FENETRE::pointilles (const CX& a, const CX& b,
  const COULEUR c, const int largeur) const
{
  unsigned int xa, ya, xb, yb;
  coordonnees_en_pixels ( xa,ya, a );
  coordonnees_en_pixels ( xb,yb, b );
  XSetForeground(dpy,gc,couleur(c));
  XGCValues gc_values;
  gc_values.line_style = LineOnOffDash;
  gc_values.line_width = largeur;
  XChangeGC (dpy,gc, GCLineWidth|GCLineStyle, &gc_values);
  XDrawLine (dpy,pix,gc, xa,ya, xb,yb);
  gc_values.line_style = LineSolid;
  gc_values.line_width = 1;
  XChangeGC (dpy,gc, GCLineWidth|GCLineStyle, &gc_values);
}

void
FENETRE::croix(const unsigned int rayon, int x, int y, const COULEUR c) const
{
  if (x<0) x += width; if (y<0) y += height;
  XSetForeground(dpy,gc,couleur(c));
  XDrawLine(dpy,pix,gc, x-rayon,y-rayon,x+rayon,y+rayon);
  XDrawLine(dpy,pix,gc, x-rayon,y+rayon,x+rayon,y-rayon);
}

void
FENETRE::cercle(const unsigned int rayon,
  int x, int y, const bool fill, const COULEUR c) const
{
  if (x<0) x += width; if (y<0) y += height;
  XSetForeground(dpy,gc,couleur(c));
  if (fill)  XFillArc(dpy,pix,gc, x-rayon,y-rayon, 2*rayon,2*rayon, 0,64*360);
  if (!fill) XDrawArc(dpy,pix,gc, x-rayon,y-rayon, 2*rayon,2*rayon, 0,64*360);
}

void
FENETRE::croix(const unsigned int rayon, const CX& a, const COULEUR c) const
{
  unsigned int xa, ya;
  coordonnees_en_pixels( xa,ya, a );
  XSetForeground(dpy,gc,couleur(c));
  XDrawLine(dpy,pix,gc, xa-rayon,ya-rayon,xa+rayon,ya+rayon);
  XDrawLine(dpy,pix,gc, xa-rayon,ya+rayon,xa+rayon,ya-rayon);
}

void
FENETRE::cercle(const unsigned int rayon,
  const CX& a, const bool fill, const COULEUR c) const
{
  unsigned int xa, ya;
  coordonnees_en_pixels( xa,ya, a );
  XSetForeground(dpy,gc,couleur(c));
  if (fill)  XFillArc(dpy,pix,gc, xa-rayon,ya-rayon, 2*rayon,2*rayon, 0,64*360);
  if (!fill) XDrawArc(dpy,pix,gc, xa-rayon,ya-rayon, 2*rayon,2*rayon, 0,64*360);
}

#include <string.h>
static const int textmaxlen = 80;
static const int textmaxnum = 20;
void
FENETRE::textaffiche (const int numero, const char* texte,
  int x, int y, const COULEUR c) const
{
  static XTextItem textitem[textmaxnum];
  static size_t    textlen [textmaxnum] = { 0, 0, 0, 0, 0 };
  static int       textx   [textmaxnum];
  static int       texty   [textmaxnum];
  static char      texts   [textmaxnum][textmaxlen+1];
  if (numero>=textmaxnum) {
    cerr << "Erreur: on ne peut afficher plus de "<<textmaxnum<<" textes\n";
    return;
  }
  textitem[numero].delta = 0;
  textitem[numero].font = None;
  if (textlen[numero]) {
    XSetForeground(dpy,gc,couleur(couleur_fond));
    XDrawText(dpy,pix,gc,textx[numero],texty[numero], textitem+numero, 1);
  }
  XSetForeground(dpy,gc,couleur(c));
  strncpy(texts[numero], texte, textmaxlen); texts[numero][textmaxlen] = '\0';
  textitem[numero].chars = texts[numero];
  textitem[numero].nchars = textlen[numero] = strlen(textitem[numero].chars);
  if (x<0) x += width; if (y<0) y += height;
  textx[numero] = x; texty[numero] = y;
  XDrawText(dpy,pix,gc,textx[numero],texty[numero], textitem+numero, 1);
}

static XEVENT
keyevent_analyse(XKeyEvent ev)
{
  XEVENT result;
  char buffer[80];
  int  buflen;
  KeySym ks;
  buflen = XLookupString(&ev, buffer,80, &ks, NULL);
  if (buflen==1) {
    result.xev_type = XEVENT::caractere;
    result.xev_char = buffer[0];
    if (ev.state & Mod1Mask) result.xev_char |= 0x80;
    return result;
  }
  if (IsModifierKey(ks)) return XEVENT ();
  result.xev_type = XEVENT::xkeysym;
  switch (ks) {
    case XK_Up:    case XK_KP_Up:     result.xev_ks = XK_Up; break;
    case XK_Down:  case XK_KP_Down:   result.xev_ks = XK_Down; break;
    case XK_Right: case XK_KP_Right:  result.xev_ks = XK_Right; break;
    case XK_Left:  case XK_KP_Left:   result.xev_ks = XK_Left; break;
    default:                          result.xev_ks = ks;
      cout << "Key sym = " << ks << "\n" << flush;
      if (buflen) printf("Key value = %.*s\n", buflen, buffer);
  }
  return result;
}

XEVENT
FENETRE::butevent_analyse (const XButtonEvent ev) const
{
  // cerr << "Bouton " << ev.button << "\n";
  return XEVENT (ev.button, coordonnees_complexes (ev.x, ev.y));
}

XEVENT
FENETRE::configure_notify ()
{
  Window wr; int xpos, ypos; unsigned int new_width, new_height, bw, depth;
  XGetGeometry(dpy,win,&wr,&xpos,&ypos, &new_width,&new_height, &bw,&depth);
  if ((width==new_width)&&(height==new_height))
    return XEVENT();
  // Changement de taille de la fentre
  scalex /= width ; width  = new_width ; scalex *= width;
  scaley /= height; height = new_height; scaley *= height;
  if (scalex < scaley) scaley = scalex; else scalex = scaley;
  XFreePixmap    (dpy,pix);
  XFreeGC        (dpy,gc);
  pix = XCreatePixmap(dpy,win, width,height,depth);
  gc  = XCreateGC(dpy,pix, 0,NULL);
  effacefenetre();
  XEVENT result;
  result.xev_type = XEVENT::redisplay;
  return result;
}

XEVENT
FENETRE::waitevent()
{
  XEvent ev;
  XNextEvent(dpy,&ev); // cout << XEventType(ev) << "\n";
  switch(ev.type) {
    case Expose:
      dessinefenetre();
    case NoExpose:
    case UnmapNotify:   // Iconification
    case MapNotify:     // D-iconification
    case MappingNotify: // Clavier remapp
      return XEVENT(); break;
    case ConfigureNotify: return configure_notify ();

    case KeyPress:     return keyevent_analyse(ev.xkey); break;
    case ButtonPress:  return butevent_analyse(ev.xbutton); break;

    default :
      cerr << "Type d'vnement inconnu " << ev.type << "\n";
      return XEVENT(ev);
  }
  cerr << "ERROR: Never Reached\n";
  return XEVENT();
}
