#ifndef H_FENETRE
#define H_FENETRE

#include "cx.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>   // XLookupString
#include <X11/keysym.h>  // Les codes XK_*

class XEVENT {
public:
  // Rajouter la souris
  enum { ignore, redisplay, caractere, bouton, xkeysym, xevent } xev_type;
  union {
    char xev_char;
    unsigned int xev_button;
    KeySym xev_ks;
    XEvent xev_xev;
  };
  CX xev_pos;
  XEVENT () { xev_type = ignore; }
  XEVENT (char c) { xev_type = caractere; xev_char = c; }
  XEVENT (unsigned int n, CX c) { xev_type=bouton; xev_button=n; xev_pos=c; }
  XEVENT (KeySym k) { xev_type = xkeysym; xev_ks = k; }
  XEVENT (XEvent x) { xev_type = xevent; xev_xev = x; }
private: // On supprime les oprateurs par dfaut
  XEVENT& operator = (const XEVENT&);
  // XEVENT (const XEVENT&);
};

struct COULEUR { int c; COULEUR(int C) : c(C) {} };
const COULEUR couleur_fond   = COULEUR(-1);
const COULEUR couleur_dessin = COULEUR(-2);

class FENETRE
{
  // lments X11
  Display *dpy;
  Pixmap pix;
  Window win;
  int scr;
  GC gc;
  unsigned int width, height;
  // chelle
  double translatex, translatey;
  double scalex, scaley;
  // Gestion des vnements
  XEVENT configure_notify ();
  XEVENT butevent_analyse (const XButtonEvent ev) const;
  // Couleur
  ulong couleur (const COULEUR c) const {
    if (c.c == couleur_fond.c) return WhitePixel(dpy,scr);
    if (c.c == couleur_dessin.c) return BlackPixel(dpy,scr);
    return c.c; }
  CX coordonnees_complexes (const unsigned int x, const unsigned int y) const;
  void coordonnees_en_pixels (unsigned int&x, unsigned int&y, const CX&c) const;
public:
  FENETRE (const char* winname);
  ~FENETRE ();
  void effacefenetre () const;
  void dessinefenetre () const;
  void axes (const COULEUR c = couleur_dessin) const;
  unsigned long RGBcolor (double r, double g, double b) const;
  void ligne (const CX& a, const CX& b,
       const COULEUR c = couleur_dessin, const int largeur = 1) const;
  void pointilles (const CX& a, const CX& b,
       const COULEUR c = couleur_dessin, const int largeur = 1) const;
  void croix (const unsigned int rayon, const CX& a,
       const COULEUR c = couleur_dessin) const;
  void croix (const unsigned int rayon, int x,int y,
       const COULEUR c = couleur_dessin) const;
  void cercle (const unsigned int rayon, const CX& a, const bool fill,
       const COULEUR c = couleur_dessin) const;
  void cercle (const unsigned int rayon, int x,int y, const bool fill,
       const COULEUR c = couleur_dessin) const;
  void textaffiche (const int numero, const char* texte,
       int x, int y, const COULEUR c = couleur_dessin) const;
  XEVENT waitevent ();
  #if defined(mplib_pari)
  int KeyPressed () const;
  #endif
  void winscale (const double factor);
  void winmove (const KeySym arrow);
private: // On supprime les oprateurs par dfaut
  FENETRE& operator = (const FENETRE&);
  FENETRE (const FENETRE&);
};

#endif
