#include "solve.h"

// On fournit x, - f(x) et f'(x)
// On calcule d tq  0 = f(x+d) = f(x) + f'(x).d 
// i.e.  d = - f'(x)^-1 . f(x);

VECTEUR
delta (const VECTEUR& x, const VECTEUR& fx, const MCARREE& der)
{
  __RE_GARBAGE_BEGIN
  // cout << x.length() << " variables et " << fx.length() << " quations\n";
  // cout << "matrice de dimension " << der.dim() << "\n";
  MCARREE i (der.dim());
  CX det = matinv (der, i);
  // cout << IOlong << "determinant = " << det << "\n" << IOshort;
  cout << IOmedium;
  cout << "determinant = " << det << "\n";
  if (IsZero(det)) {
    __RE_GARBAGE_END
    throw nulldet;
  }
  VECTEUR delta = i * fx;
  // cout << "position : " << x;
  // cout << "dcalage : " << delta;
#if defined(mplib_pari)
  ulong dim = delta.length();
  GEN *gptr [2*dim];
  for (ulong j = 0; j < dim; j++)
      genptr (gptr, j, delta[j]);
  gerepilemany(avma_garbage, gptr, 2*dim);
#endif
  return delta;
}

VECTEUR
dichosolve (const VECTEUR& x, const VECTEUR& fx, const MCARREE& der,
       VECTEUR (*f) (const VECTEUR& x) )
{
  VECTEUR d = delta(x, fx, der);
  RE oldnorm = norm (fx);
    cout << "OldNorm = " << oldnorm << "\n";
  RE newnorm (-1);
  long loop = -1;
  do {
    loop++;
    if (loop > 9) throw dicho;
    if (newnorm > RE(-1)) d /= 2;
    VECTEUR newx = x; newx += d;
    VECTEUR newfx = f(newx);
    newnorm = norm (newfx);
    cout << "NewNorm = " << newnorm << "\n";
  } while (newnorm > oldnorm);
  cout << "Delta / 2^" << loop << "\n";
  return d;
}

static CX
evalue (const POLYNOME& p, const CX& val)
{
  CX res = p[p.degre()];
  for (ulong i = p.degre(); i > 0; i--) { res *= val; res += p[i-1]; }
  return res;
}

static CX
racine (const POLYNOME& p, const CX& estimation)
{
  if (p.degre() == 0) return estimation;
  if (p.degre() == 1) return - p[0]/p[1];
  CX x = estimation;
  for (ulong i = 0; i < 20; i++) {
    CX px = evalue (p, x);
    if (IsZero (px)) break;
    POLYNOME q = p; q.derivee();
    CX qx = evalue (q, x);
    if (IsZero (qx)) return estimation; // problme
    if (i < 5) x -= px / qx / 2;
    else       x -= px / qx;
  }
  return x;
}

VECTEUR
racines (const POLYNOME& p, const VECTEUR& estimation)
{
  bool* oksol = new bool[estimation.length()];
  {for (ulong i = 0; i < estimation.length(); i++) oksol[i] = false;}
  VECTEUR solution (estimation.length());
  POLYNOME q = p;
  for (ulong i = 0; i < estimation.length(); i++) {
    if (oksol[i]) continue;
    solution[i] = racine (q, estimation[i]);
    CX v = evalue(q, solution[i]);
    oksol[i] = NearZero (v);
    // cerr << "Racines[" << i << "] : " << IOpari << v
    //      << " [" << 2 * log2(v.Re()) << ">" << - precision(v.Re()) << "]\n";
    if (!oksol[i]) continue;
    // la plus proche estimation de solution[i]
    ulong k = i;
    for (ulong j = 0; j < estimation.length(); j++) {
      if (oksol[j]) continue;
      if (norm(solution[i]-estimation[j]) > norm(solution[i]-estimation[k]))
        continue;
      k = j;
    }
    if (k != i) {
      // cout << "Swap[" << i << "," << k << "]\n";
      solution[k] = solution[i];
      oksol[i] = false;
      oksol[k] = true;
    }
    q.div_xan(solution[i], 1);
  }
  for (ulong i = 0; i < estimation.length(); i++) {
    if (oksol[i]) continue;
    solution[i] = racine (q, estimation[i]);
    CX v = evalue(q, solution[i]);
    oksol[i] = NearZero (v);
    if (!oksol[i]) {
      cerr << "Racines[" << i << "] : imprcision " << IOpari << v
           << " [" << 2 * log2(v.Re()) << ">" << - precision(v.Re()) << "]\n";
    }
    q.div_xan(solution[i], 1);
  }
  return solution;
}
